/*=========================================================================

  Program:   Insight Segmentation & Registration Toolkit
  Module:    $RCSfile: evaluator.h,v $
  Language:  C++
  Date:      $Date: 2004/08/06 12:38:33 $
  Version:   $Revision: 1.35 $

  Copyright (c) Insight Software Consortium. All rights reserved.
  See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __evaluator_h
#define __evaluator_h


#include <cmath>
#include <vector>
#include <valarray>



#include "polynomial.h"
#include "macros.h"



namespace fmi
{

   
class evaluator
{
   public:

      //typedef float ScalarType;
      
      /** Set the source landmarks list. Since this is 3D, landmarks must *
      point to vector containing 3*numberOfPoints elements. The *
      numberOfPoints must be the same as passed to the next function.*/
      void SetPointSet(const ScalarType* const landmarks,
		       const int numberOfPoints);
  
      /** Set the coefficients corresponding to the landmarks. The
      * numberOfPoints must be the same as passed to the previous function.
      * Note: this vector includes the extra polynomial coefficients.  Thus if
      * the degree of the extra polynomial is 1 and there are 1000 points then
      * coefficients must point to a vector containing 1004 elements. */
      void SetCoefficients(const ScalarType* const coefficients, 
			   const int numberOfPoints);


      void SetExtraPolynomial(int degree,const ScalarType* extra);
      //void SetExtraPolynomialDegree(int d);
      int GetExtraPolynomialDegree() const { return m_ExtraPolynomialDegree; }

      /** Default is 2 */
      void SetDepth(int depth);

      /** Set the depth to what the algorithm calculates to be the
      * best to use */ 
      void SetDepthToOptimal();
   
      /** Default is 5 */
      void SetP(int p);

      /** Default is 1 */
      void SetK(int k);

      /** Default is 0.0 */
      void SetTau(ScalarType t);

      void SetError(ScalarType e);
      ScalarType GetError() const {return m_Error;}

      /** Default is a bounding box of all the source landmarks
      * plus a small border.
      * You can optionaly make it bigger.
      * If you don't set it, it will be computed automatically */
      void SetBoundingBox(const ScalarType *box);

      // The following function determines how we go about determining whether
      // a point is well seperated from a panel.
      void SetWellSeparatedFactor(ScalarType w) {m_WellSeparatedFactor2=w*w;}

      // The following functions determine how we decrease the degree of the
      // polynomial used as we get farther and farther away from the center of
      // the panel. 
      void SetDecreaseFactor(ScalarType d) {m_DecreaseFactor=d;}
      void SetDecreaseFactor2(ScalarType d) {m_DecreaseFactor2=d;}
      void SetMinDegree(int m) {m_MinDegree=m;}

      /** Call this function after setting up the class.  It will compute the
      * fast multipole polynomials.  If you modify anything then call this
      * again and it will be updated. */
      void Update();


      /** This function should only be called after calling Update. */
      ScalarType Evaluate( const ScalarType input[3] ) const;

      template<class PointType>
      ScalarType EvaluateT( const PointType& input ) const
      {	return Evaluate(&input[0]); }
   
      /** This function should only be called after calling Update. */
      void EvaluateAtAllSourcePoints(ScalarType* const result) const;
      void EvaluateAtAllSourcePointsNew(ScalarType* const result) const;

      /** For testing purposes. */
      ScalarType EvaluateDirect( const ScalarType input[3] ) const;
      void EvaluateAtAllSourcePointsDirect(ScalarType* const result) const;


      /** The following will evaluate the sum over a grid using the FMM. You
      * specify the grid by giving the spacing, origin and size of the
      * grid. No checking is done to make sure the grid is contained in the
      * bounding box. */ 
      void EvaluateGrid( ScalarType* result, 
			 const ScalarType spacing[3],
			 const ScalarType origin[3],
			 const int size[3]) const;

      /** For testing purposes */ 
/*      void EvaluateGridDirect( ScalarType* result, 
			       const ScalarType spacing[3],
			       const ScalarType origin[3],
			       const int size[3]) const;
*/
  

      // For testing purposes
      void test();

      evaluator();
      ~evaluator() {}

   protected:


      typedef Polynomial PolynomialType;

      void UpwardPass();


      void PolynomialGenerator(PolynomialType &G,
			       const ScalarType t[4],
			       const ScalarType coef) const;

      ScalarType BasicFunction(const ScalarType point[3], 
			       const ScalarType center[3]) const;

      ScalarType EvaluateSumBasicFunctions( const ScalarType point[3],
					    const int box ) const;

      void EvaluateSumBasicFunctionsNew( const ScalarType point[3],
					 const int* const boxToPointMap,
					 ScalarType& result) const;

      ScalarType EvaluateExtraPolynomialTerms( 
	 const ScalarType thisPoint[3] ) const;

/*
      ScalarType EvaluateFarFieldPolynomial2( const ScalarType point[4],
					     const PolynomialType& poly,
					     const unsigned int maxDegree) const;
*/


      void EvaluateFarFieldPolynomial( const ScalarType point[3],
				       const PolynomialType& poly,
				       const unsigned int maxDegree,
				       ScalarType& result) const;

	 
      void ComputeBoundingBox();

      void SortPoints();

      void SetupOctree();

      class PointType
      {
	    ScalarType p[3];
	 public:
	    ScalarType& operator[] (unsigned int i)
	    {
	       return p[ i ];
	    }
	    ScalarType const & operator[] (unsigned int i) const
	    {
	       return p[ i ];
	    }
      };




      //typedef std::vector< std::pair<int, pointvalue> > PointContainerType;
      //typedef std::vector< PointType > PointContainerType;
      //PointContainerType m_Points;

      std::vector< PointType > m_Points;
      std::valarray< ScalarType > m_Coefficients;
      
      
      // This hold the coefficients of an extra polynomial. There are 4
      // supported polynomials: None, constant, linear, quadratic
      std::valarray< ScalarType > m_ExtraPolynomialCoefficients;

      int m_NumberOfPoints;

      ScalarType m_BoundingBox[6];

      ScalarType m_Tau;
      ScalarType m_Tau2;

      //bool m_PointsAreSorted;
      //bool m_PointsAreInsideUnitCube;
      int m_Depth;
      int m_P;
      int m_K;
      int m_PplusK;
      int m_ExtraPolynomialDegree;
      //int m_NumberExtraPolynomialTerms;
      
      typedef std::vector<PolynomialType> PolynomialArrayType;
      std::vector<PolynomialArrayType> m_FarFieldExpansions;

      std::valarray<int> m_NumberOfBoxesPerLevel;

      std::valarray<ScalarType> m_A; // See Eq. 5.4 in paper
      std::valarray<ScalarType> m_B; // See Eq. 5.4 in paper

      // This array holds the centers of the boxes at each level
      std::vector< std::vector<PointType> > m_CenterPerBox;

      // In this array, the ith element contains an index into the *sorted*
      // pointset of the first point in the ith box

      std::vector< std::valarray<int> > m_BoxToPointMap;
      //std::valarray<int> m_BoxToPointMap;

      // This holds for each level, the box containing the ith point
      std::vector< std::valarray<int> > m_PointToBoxMap;
      //std::valarray<int> m_PointToBoxMap;

      std::valarray<int> m_OriginalOrderOfPoints;



      ScalarType m_WellSeparatedFactor2;
      ScalarType m_DecreaseFactor;
      ScalarType m_DecreaseFactor2;
      int m_MinDegree;
      ScalarType m_Error;
      std::valarray<int> m_NumberCutoffPoints;
//      std::vector< std::valarray<ScalarType> > m_M;
      std::vector< std::valarray<ScalarType> > m_R2;
//      std::vector< std::valarray<ScalarType> > m_Error2KMRK;
//      std::vector< std::vector<ScalarType> > m_M;
//      std::valarray<ScalarType> m_R;
//      std::valarray<ScalarType> m_Error2KRK;



      // Flags for determining what to update when Update() is called
      bool m_Modified;
      bool m_ModifiedCoefficients;


   private:
      evaluator(const evaluator&); //purposely not implemented
      void operator=(const evaluator&); //purposely not implemented

      //mutable PolynomialType m_Tmp;
};







   inline
   ScalarType 
   evaluator::
   BasicFunction(const ScalarType point[3], 
		 const ScalarType center[3]) const
   {
      const ScalarType x = point[0] - center[0];
      const ScalarType y = point[1] - center[1];
      const ScalarType z = point[2] - center[2];
      //return pow(x*x + y*y + z*z + m_Tau*m_Tau, (ScalarType)m_K/2.0 );
//	 adds+=6;mults+=3;pows++;sqrts++;
      return sqrt( std::pow(x*x + y*y + z*z + m_Tau2, m_K ) ); 
   }
   
   inline 
   ScalarType
   evaluator::
   EvaluateSumBasicFunctions( const ScalarType point[3],
			      const int box) const
   {
      
      ScalarType result = 0.0;

      const int indexEnd = m_BoxToPointMap[m_Depth][box+1];

      //const pointvalue* ptr = &m_PointsValues[ m_BoxToPointMap[box] ];

      for(int i=m_BoxToPointMap[m_Depth][box]; i<indexEnd; ++i)
      {

	 result += m_Coefficients[i] *
	    this->BasicFunction(point,&m_Points[i][0]);
	 //result += ptr->v *
	 //   this->BasicFunction(point,&(*ptr)[0]);
	 //++ptr;
      }

      return result;
   }


   inline 
   void
   evaluator::
   EvaluateSumBasicFunctionsNew( const ScalarType point[3],
				 const int* const boxToPointMap,
				 ScalarType& result) const
   {
      
      //ScalarType result = 0.0;

      const int indexEnd = boxToPointMap[1];

      for(int i=boxToPointMap[0]; i<indexEnd; ++i)
      {

	 result += m_Coefficients[i] *
	    this->BasicFunction(point,&m_Points[i][0]);

      }


   }






}


#endif 
